"""
Mypy plugin.

This is a simple plugin, that adds support for our custom attrs wrapper.

To use it, add `interactions.ext.mypy` to the plugins list in your mypy config.

For mypy.ini:
```ini
[mypy]
plugins = interactions.ext.mypy
```

For pyproject.toml:
```toml
[tool.mypy]
plugins = "interactions.ext.mypy"
```
"""
from functools import partial
from typing import Callable, Optional, Type
from mypy.plugin import Plugin, ClassDefContext
from mypy.plugins import attrs

__all__ = ("plugin",)


class NaffPlugin(Plugin):
    # This could be smarter, but it does the job.
    def get_class_decorator_hook(self, fullname: str) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname == "interactions.client.utils.attr_utils.define":
            return partial(
                attrs.attr_class_maker_callback,
                auto_attribs_default=None,
            )
        return None


def plugin(version: str) -> Type[NaffPlugin]:
    # ignore version argument if the plugin works with all mypy versions.
    return NaffPlugin



import io
import textwrap
import traceback
from contextlib import redirect_stdout
from typing import Any, Optional

from interactions import (
    Extension,
    slash_command,
    InteractionContext,
    Modal,
    ParagraphText,
    MISSING,
    ModalContext,
)
from interactions.ext.debug_extension.utils import debug_embed
from interactions.ext.paginators import Paginator
from interactions.models import (
    Embed,
    Message,
    File,
)

__all__ = ("DebugExec",)


class DebugExec(Extension):
    def __init__(self, bot) -> None:
        self.cache: dict[int, str] = {}

    @slash_command("debug", sub_cmd_name="exec", sub_cmd_description="Run arbitrary code")
    async def debug_exec(self, ctx: InteractionContext) -> Optional[Message]:
        last_code = self.cache.get(ctx.author.id, MISSING)

        env = {
            "bot": self.bot,
            "ctx": ctx,
            "channel": ctx.channel,
            "author": ctx.author,
            "server": ctx.guild,
            "guild": ctx.guild,
            "message": ctx.message,
        } | globals()

        modal = Modal(
            title="Debug-Exec",
            components=[
                ParagraphText(
                    label="Code to run",
                    value=last_code,
                    custom_id="body",
                    placeholder="Write your code here!",
                )
            ],
        )
        await ctx.send_modal(modal)

        m_ctx = await self.bot.wait_for_modal(modal, ctx.author)
        await m_ctx.defer()

        body = m_ctx.kwargs["body"]
        self.cache[ctx.author.id] = body

        if body.startswith("```") and body.endswith("```"):
            body = "\n".join(body.split("\n")[1:-1])
        else:
            body = body.strip("` \n")

        stdout = io.StringIO()

        to_compile = "async def func():\n%s" % textwrap.indent(body, "  ")
        try:
            exec(to_compile, env)
        except SyntaxError:
            return await ctx.send(f"```py\n{traceback.format_exc()}\n```")

        func = env["func"]
        try:
            with redirect_stdout(stdout):
                ret = await func()
        except Exception:
            return await m_ctx.send(f"```py\n{stdout.getvalue()}{traceback.format_exc()}\n```")
        else:
            return await self.handle_exec_result(m_ctx, ret, stdout.getvalue(), body)

    async def handle_exec_result(self, ctx: ModalContext, result: Any, value: Any, body: str) -> Optional[Message]:
        # body can be of length 2000 and exceed the limit after formatting
        if len(cmd_body := f"```py\n{body}```") <= 2000:
            await ctx.send(cmd_body)

        else:
            paginator = Paginator.create_from_string(self.bot, body, prefix="```py", suffix="```", page_size=4000)
            await paginator.send(ctx)

        if result is None:
            result = value or "No Output!"

        if isinstance(result, Message):
            try:
                e = debug_embed("Exec", timestamp=result.created_at, url=result.jump_url)
                e.description = result.content
                e.set_author(
                    result.author.tag,
                    icon_url=(result.author.guild_avatar or result.author.avatar).url,
                )
                e.add_field("\u200b", f"[Jump To]({result.jump_url})\n{result.channel.mention}")

                return await ctx.send(embeds=e)
            except Exception:
                return await ctx.send(result.jump_url)

        if isinstance(result, Embed):
            return await ctx.send(embeds=[result])

        if isinstance(result, File):
            return await ctx.send(file=result)

        if isinstance(result, Paginator):
            return await result.send(ctx)

        if hasattr(result, "__iter__"):
            l_result = list(result)
            if all(isinstance(r, Embed) for r in result):
                paginator = Paginator.create_from_embeds(self.bot, *l_result)
                return await paginator.send(ctx)

        if not isinstance(result, str):
            result = repr(result)

        # prevent token leak
        result = result.replace(self.bot.http.token, "[REDACTED TOKEN]")

        if len(cmd_result := f"```py\n{result}```") <= 2000:
            return await ctx.send(cmd_result)

        paginator = Paginator.create_from_string(self.bot, result, prefix="```py", suffix="```", page_size=4000)
        return await paginator.send(ctx)


def setup(bot) -> None:
    DebugExec(bot)



import re
import interactions.models as models

__all__ = ("mentions",)


def mentions(
    text: str,
    query: "str | re.Pattern[str] | models.BaseUser | models.BaseChannel | models.Role",
    *,
    tag_as_mention: bool = False,
) -> bool:
    """
    Checks whether a query is present in a text.

    Args:
        text: The text to search in
        query: The query to search for
        tag_as_mention: Should `BaseUser.tag` be checked *(only if query is an instance of BaseUser)*

    Returns:
        Whether the query could be found in the text
    """
    if isinstance(query, str):
        return query in text
    if isinstance(query, re.Pattern):
        return query.match(text) is not None
    if isinstance(query, models.BaseUser):
        # mentions with <@!ID> aren't detected without the replacement
        return (query.mention in text.replace("@!", "@")) or (query.tag in text if tag_as_mention else False)
    if isinstance(query, (models.BaseChannel, models.Role)):
        return query.mention in text
    return False



from interactions import Snowflake_Type, Client, Message
from interactions.models.internal.context import InteractionContext
from tests.consts import SAMPLE_MESSAGE_DATA
from interactions.ext.prefixed_commands import PrefixedContext

__all__ = ("generate_dummy_context",)


def generate_dummy_context(
    user_id: Snowflake_Type | None = None,
    channel_id: Snowflake_Type | None = None,
    guild_id: Snowflake_Type | None = None,
    message_id: Snowflake_Type | None = None,
    dm: bool = False,
    client: Client | None = None,
) -> InteractionContext:
    """Generates a dummy context for testing."""
    client = Client() if client is None else client

    if not dm and not guild_id:
        guild_id = "123456789012345670"
    elif dm:
        guild_id = None

    # channel = SAMPLE_CHANNEL_DATA(channel_id=channel_id, guild_id=guild_id)
    message = SAMPLE_MESSAGE_DATA(user_id=user_id, channel_id=channel_id, message_id=message_id, guild_id=guild_id)

    return PrefixedContext.from_message(client, Message.from_dict(message, client))



import time
from collections import OrderedDict
from collections.abc import ItemsView, ValuesView
from typing import Any, Callable, Generic, Iterator, Optional, Tuple, TypeVar

import attrs

__all__ = ("TTLItem", "TTLCache", "NullCache")

KT = TypeVar("KT")
VT = TypeVar("VT")


class NullCache(dict):
    """
    A special cache that will always return None

    Effectively just a lazy way to disable caching.
    """

    def __setitem__(self, key, value) -> None:
        pass


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class TTLItem(Generic[VT]):
    value: VT = attrs.field(
        repr=False,
    )
    expire: float = attrs.field(
        repr=False,
    )
    """When the item expires in cache."""

    def is_expired(self, timestamp: float) -> bool:
        """
        Check if the item is expired.

        Args:
            timestamp: The current timestamp to compare against.

        Returns:
            True if the item is expired, False otherwise.

        """
        return timestamp >= self.expire


class TTLCache(OrderedDict[KT, TTLItem[VT]]):
    def __init__(
        self,
        ttl: int = 600,
        soft_limit: int = 50,
        hard_limit: int = 250,
        on_expire: Optional[Callable] = None,
    ) -> None:
        super().__init__()

        self.ttl = ttl
        self.hard_limit = hard_limit
        self.soft_limit = min(soft_limit, hard_limit)
        self.on_expire = on_expire

    def __setitem__(self, key: KT, value: VT) -> None:
        expire = time.monotonic() + self.ttl
        item = TTLItem(value, expire)
        super().__setitem__(key, item)
        self.move_to_end(key)

        self.expire()

    def __getitem__(self, key: KT) -> VT:
        # Will not (should not) reset expiration!
        item = super().__getitem__(key)
        # self._reset_expiration(key, item)
        return item.value

    def pop(self, key: KT, default=attrs.NOTHING) -> VT:
        if key in self:
            item = self[key]
            del self[key]
            return item

        if default is attrs.NOTHING:
            raise KeyError(key)

        return default

    def get(self, key: KT, default: Optional[VT] = None, reset_expiration: bool = True) -> VT:
        item = super().get(key, default)
        if item is not default:
            if reset_expiration:
                self._reset_expiration(key, item)
            return item.value

        return default

    def values(self) -> ValuesView[VT]:
        return _CacheValuesView(self)

    def items(self) -> ItemsView:
        return _CacheItemsView(self)

    def _reset_expiration(self, key: KT, item: TTLItem) -> None:
        self.move_to_end(key)
        item.expire = time.monotonic() + self.ttl

    def _first_item(self) -> Tuple[KT, TTLItem[VT]]:
        return next(super().items().__iter__())

    def expire(self) -> None:
        """Removes expired elements from the cache."""
        if self.soft_limit and len(self) <= self.soft_limit:
            return

        if self.hard_limit:
            while len(self) > self.hard_limit:
                self._expire_first()

        timestamp = time.monotonic()
        while True:
            key, item = self._first_item()
            if item.is_expired(timestamp):
                self._expire_first()
            else:
                break

    def _expire_first(self) -> None:
        key, value = self.popitem(last=False)
        if self.on_expire:
            self.on_expire(key, value)


class _CacheValuesView(ValuesView):
    def __contains__(self, value) -> bool:
        for key in self._mapping:
            v = self._mapping.get(key, reset_expiration=False)
            if v is value or v == value:
                return True
        return False

    def __iter__(self) -> Iterator[Any]:
        for key in self._mapping:
            yield self._mapping.get(key, reset_expiration=False)

    def __reversed__(self) -> Iterator[Any]:
        for key in reversed(self._mapping):
            yield self._mapping.get(key, reset_expiration=False)


class _CacheItemsView(ItemsView):
    def __contains__(self, item) -> bool:
        key, value = item
        v = self._mapping.get(key, default=attrs.NOTHING, reset_expiration=False)
        return False if v is attrs.NOTHING else v is value or v == value

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        for key in self._mapping:
            yield key, self._mapping.get(key, reset_expiration=False)

    def __reversed__(self) -> Iterator[Tuple[Any, Any]]:
        for key in reversed(self._mapping):
            yield key, self._mapping.get(key, reset_expiration=False)



import asyncio
import inspect
from asyncio import Task as _Task
from datetime import datetime, timedelta
from typing import Callable

import interactions
from interactions.client.const import get_logger
from .triggers import BaseTrigger

__all__ = ("Task",)


class Task:
    """
    Create an asynchronous background tasks. Tasks allow you to run code according to a trigger object.

    A task's trigger must inherit from `BaseTrigger`.

    Attributes:
        callback (Callable): The function to be called when the trigger is triggered.
        trigger (BaseTrigger): The trigger object that determines when the task should run.
        task (Optional[_Task]): The task object that is running the trigger loop.
        iteration (int): The number of times the task has run.

    """

    callback: Callable
    trigger: BaseTrigger
    task: _Task | None
    _stop: asyncio.Event
    iteration: int

    def __init__(self, callback: Callable, trigger: BaseTrigger) -> None:
        self.callback = callback
        self.trigger = trigger
        self._stop = asyncio.Event()
        self.task = None
        self.iteration = 0

    @property
    def started(self) -> bool:
        """Whether the task is started"""
        return self.task is not None

    @property
    def running(self) -> bool:
        """Whether the task is running"""
        return self.task is not None and not self.task.done()

    @property
    def done(self) -> bool:
        """Whether the task is done/finished"""
        return self.task is not None and self.task.done()

    @property
    def next_run(self) -> datetime | None:
        """Get the next datetime this task will run."""
        return self.trigger.next_fire() if self.running else None

    @property
    def delta_until_run(self) -> timedelta | None:
        """Get the time until the next run of this task."""
        if not self.running:
            return None

        next_run = self.next_run
        return next_run - datetime.now() if next_run is not None else None

    def on_error_sentry_hook(self, error: Exception) -> None:
        """A dummy method for interactions.ext.sentry to hook"""

    def on_error(self, error: Exception) -> None:
        """Error handler for this task. Called when an exception is raised during execution of the task."""
        self.on_error_sentry_hook(error)
        interactions.Client.default_error_handler("Task", error)

    async def __call__(self) -> None:
        try:
            if inspect.iscoroutinefunction(self.callback):
                val = await self.callback()
            else:
                val = self.callback()

            if isinstance(val, BaseTrigger):
                self.reschedule(val)
        except Exception as e:
            self.on_error(e)

    def _fire(self, fire_time: datetime) -> None:
        """Called when the task is being fired."""
        self.trigger.last_call_time = fire_time
        _ = asyncio.create_task(self())
        self.iteration += 1

    async def _task_loop(self) -> None:
        """The main task loop to fire the task at the specified time based on triggers configured."""
        while not self._stop.is_set():
            fire_time = self.trigger.next_fire()
            if fire_time is None:
                return self.stop()

            future = asyncio.create_task(self._stop.wait())
            timeout = (fire_time - datetime.now()).total_seconds()
            done, _ = await asyncio.wait([future], timeout=timeout, return_when=asyncio.FIRST_COMPLETED)
            if future in done:
                return None

            self._fire(fire_time)

    def start(self) -> None:
        """Start this task."""
        try:
            self.trigger.reschedule()
            self._stop.clear()
            self.task = asyncio.create_task(self._task_loop())
        except RuntimeError:
            get_logger().error(
                "Unable to start task without a running event loop! We recommend starting tasks within an `on_startup` event."
            )

    def stop(self) -> None:
        """End this task."""
        self._stop.set()
        if self.task:
            self.task.cancel()

    def restart(self) -> None:
        """Restart this task."""
        self.stop()
        self.start()

    def reschedule(self, trigger: BaseTrigger) -> None:
        """
        Change the trigger being used by this task.

        Args:
            trigger: The new Trigger to use

        """
        self.trigger = trigger
        self.restart()

    @classmethod
    def create(cls, trigger: BaseTrigger) -> Callable[[Callable], "Task"]:
        """
        A decorator to create a task.

        Example:
            ```python
            @Task.create(IntervalTrigger(minutes=5))
            async def my_task():
                print("It's been 5 minutes!")

            @listen()
            async def on_startup():
                my_task.start()
            ```

        Args:
            trigger: The trigger to use for this task

        """

        def wrapper(func: Callable) -> "Task":
            return cls(func, trigger)

        return wrapper



from typing import TYPE_CHECKING, cast

import discord_typings

from interactions.models.internal.protocols import CanRequest
from ..route import Route

__all__ = ("EmojiRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type


class EmojiRequests(CanRequest):
    async def get_all_guild_emoji(self, guild_id: "Snowflake_Type") -> list[discord_typings.EmojiData]:
        """
        Get all the emoji from a guild.

        Args:
            guild_id: The ID of the guild to query.

        Returns:
            List of emoji objects

        """
        result = await self.request(Route("GET", f"/guilds/{int(guild_id)}/emojis"))
        return cast(list[discord_typings.EmojiData], result)

    async def get_guild_emoji(
        self, guild_id: "Snowflake_Type", emoji_id: "Snowflake_Type"
    ) -> discord_typings.EmojiData:
        """
        Get a specific guild emoji object.

        Args:
            guild_id: The ID of the guild to query
            emoji_id: The ID of the emoji to get

        Returns:
            PartialEmoji object

        """
        result = await self.request(Route("GET", f"/guilds/{int(guild_id)}/emojis/{int(emoji_id)}"))
        return cast(discord_typings.EmojiData, result)

    async def create_guild_emoji(
        self, payload: dict, guild_id: "Snowflake_Type", reason: str | None = None
    ) -> discord_typings.EmojiData:
        """
        Create a guild emoji.

        Args:
            payload: The emoji's data
            guild_id: The ID of the guild
            reason: The reason for this creation

        Returns:
            The created emoji object

        """
        result = await self.request(Route("POST", f"/guilds/{int(guild_id)}/emojis"), payload=payload, reason=reason)
        return cast(discord_typings.EmojiData, result)

    async def modify_guild_emoji(
        self,
        payload: dict,
        guild_id: "Snowflake_Type",
        emoji_id: "Snowflake_Type",
        reason: str | None = None,
    ) -> discord_typings.EmojiData:
        """
        Modify an existing guild emoji.

        Args:
            payload: The emoji's data
            guild_id: The ID of the guild
            emoji_id: The ID of the emoji to update
            reason: The reason for this creation

        Returns:
            The updated emoji object

        """
        result = await self.request(
            Route("PATCH", f"/guilds/{int(guild_id)}/emojis/{int(emoji_id)}"),
            payload=payload,
            reason=reason,
        )
        return cast(discord_typings.EmojiData, result)

    async def delete_guild_emoji(
        self, guild_id: "Snowflake_Type", emoji_id: "Snowflake_Type", reason: str | None = None
    ) -> None:
        """
        Delete a guild emoji.

        Args:
            guild_id: The ID of the guild
            emoji_id: The ID of the emoji to update
            reason: The reason for this deletion

        """
        await self.request(Route("DELETE", f"/guilds/{int(guild_id)}/emojis/{int(emoji_id)}"), reason=reason)



from .annotations import (
    slash_attachment_option,
    slash_bool_option,
    slash_channel_option,
    slash_float_option,
    slash_int_option,
    slash_mentionable_option,
    slash_role_option,
    slash_str_option,
    slash_user_option,
)
from .callback import CallbackObject
from .active_voice_state import ActiveVoiceState
from .application_commands import (
    application_commands_to_dict,
    auto_defer,
    CallbackType,
    component_callback,
    ComponentCommand,
    context_menu,
    user_context_menu,
    message_context_menu,
    ContextMenu,
    global_autocomplete,
    GlobalAutoComplete,
    InteractionCommand,
    LocalisedDesc,
    LocalisedName,
    LocalizedDesc,
    LocalizedName,
    modal_callback,
    ModalCommand,
    OptionType,
    slash_command,
    slash_default_member_permission,
    slash_option,
    SlashCommand,
    SlashCommandChoice,
    SlashCommandOption,
    SlashCommandParameter,
    subcommand,
    sync_needed,
)
from .auto_defer import AutoDefer
from .checks import dm_only, guild_only, has_any_role, has_id, has_role, is_owner
from .command import BaseCommand, check, cooldown, max_concurrency
from .context import (
    AutocompleteContext,
    BaseContext,
    BaseInteractionContext,
    ComponentContext,
    ContextMenuContext,
    InteractionContext,
    ModalContext,
    Resolved,
    SlashContext,
)
from .converters import (
    BaseChannelConverter,
    ChannelConverter,
    CustomEmojiConverter,
    DMChannelConverter,
    DMConverter,
    DMGroupConverter,
    Greedy,
    GuildCategoryConverter,
    GuildChannelConverter,
    GuildConverter,
    GuildNewsConverter,
    GuildNewsThreadConverter,
    GuildPrivateThreadConverter,
    GuildPublicThreadConverter,
    GuildStageVoiceConverter,
    GuildTextConverter,
    GuildVoiceConverter,
    IDConverter,
    MemberConverter,
    MessageableChannelConverter,
    MessageConverter,
    MODEL_TO_CONVERTER,
    NoArgumentConverter,
    PartialEmojiConverter,
    RoleConverter,
    SnowflakeConverter,
    ThreadChannelConverter,
    UserConverter,
    VoiceChannelConverter,
)
from .cooldowns import Buckets, Cooldown, CooldownSystem, MaxConcurrency
from .listener import listen, Listener
from .protocols import Converter
from .extension import Extension
from .wait import Wait
from .tasks import BaseTrigger, DateTrigger, IntervalTrigger, OrTrigger, Task, TimeTrigger

__all__ = (
    "ActiveVoiceState",
    "application_commands_to_dict",
    "auto_defer",
    "AutocompleteContext",
    "AutoDefer",
    "BaseChannelConverter",
    "BaseCommand",
    "BaseContext",
    "BaseInteractionContext",
    "BaseTrigger",
    "Buckets",
    "CallbackObject",
    "CallbackType",
    "ChannelConverter",
    "check",
    "component_callback",
    "ComponentCommand",
    "ComponentContext",
    "context_menu",
    "user_context_menu",
    "message_context_menu",
    "ContextMenu",
    "ContextMenuContext",
    "Converter",
    "cooldown",
    "Cooldown",
    "CooldownSystem",
    "CustomEmojiConverter",
    "DateTrigger",
    "dm_only",
    "DMChannelConverter",
    "DMConverter",
    "DMGroupConverter",
    "Extension",
    "global_autocomplete",
    "GlobalAutoComplete",
    "Greedy",
    "guild_only",
    "GuildCategoryConverter",
    "GuildChannelConverter",
    "GuildConverter",
    "GuildNewsConverter",
    "GuildNewsThreadConverter",
    "GuildPrivateThreadConverter",
    "GuildPublicThreadConverter",
    "GuildStageVoiceConverter",
    "GuildTextConverter",
    "GuildVoiceConverter",
    "has_any_role",
    "has_id",
    "has_role",
    "IDConverter",
    "InteractionCommand",
    "InteractionContext",
    "IntervalTrigger",
    "is_owner",
    "listen",
    "Listener",
    "LocalisedDesc",
    "LocalisedName",
    "LocalizedDesc",
    "LocalizedName",
    "max_concurrency",
    "MaxConcurrency",
    "MemberConverter",
    "MessageableChannelConverter",
    "MessageConverter",
    "modal_callback",
    "ModalCommand",
    "ModalContext",
    "MODEL_TO_CONVERTER",
    "NoArgumentConverter",
    "OptionType",
    "OrTrigger",
    "PartialEmojiConverter",
    "Resolved",
    "RoleConverter",
    "slash_attachment_option",
    "slash_bool_option",
    "slash_channel_option",
    "slash_command",
    "slash_default_member_permission",
    "slash_float_option",
    "slash_int_option",
    "slash_mentionable_option",
    "slash_option",
    "slash_role_option",
    "slash_str_option",
    "slash_user_option",
    "SlashCommand",
    "SlashCommandChoice",
    "SlashCommandOption",
    "SlashCommandParameter",
    "SlashContext",
    "SnowflakeConverter",
    "subcommand",
    "sync_needed",
    "Task",
    "ThreadChannelConverter",
    "TimeTrigger",
    "UserConverter",
    "VoiceChannelConverter",
    "Wait",
)



from typing import Any, TYPE_CHECKING, List

import discord_typings

from interactions.client.const import MISSING, Absent
from ..route import Route

__all__ = ("ScheduledEventsRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type


class ScheduledEventsRequests:
    request: Any

    async def list_schedules_events(
        self, guild_id: "Snowflake_Type", with_user_count: bool = False
    ) -> List[discord_typings.GuildScheduledEventData]:
        """
        Get the scheduled events for a guild.

        Args:
            guild_id: The guild to get scheduled events from
            with_user_count: Whether to include the user count in the response
        Returns:
            List of Scheduled Events or None

        """
        return await self.request(
            Route("GET", f"/guilds/{guild_id}/scheduled-events?with_user_count={with_user_count}")
        )

    async def get_scheduled_event(
        self,
        guild_id: "Snowflake_Type",
        scheduled_event_id: "Snowflake_Type",
        with_user_count: bool = False,
    ) -> discord_typings.GuildScheduledEventData:
        """
        Get a scheduled event for a guild.

        Args:
            guild_id: The guild to get scheduled event from
            scheduled_event_id: The target scheduled event to fetch
            with_user_count: Whether to include the user count in the response

        Returns:
            Scheduled Event or None

        """
        return await self.request(
            Route("GET", f"/guilds/{guild_id}/scheduled-events/{scheduled_event_id}"),
            params={"with_user_count": with_user_count},
        )

    async def create_scheduled_event(
        self,
        guild_id: "Snowflake_Type",
        payload: dict,
        reason: Absent[str] = MISSING,
    ) -> discord_typings.GuildScheduledEventData:
        """
        Create a scheduled event for a guild.

        Args:
            guild_id: The guild to create scheduled event from
            payload: The scheduled event payload
            reason: The reason to be displayed in audit logs

        Returns:
            Scheduled Event or None

        """
        return await self.request(Route("POST", f"/guilds/{guild_id}/scheduled-events"), payload=payload, reason=reason)

    async def modify_scheduled_event(
        self,
        guild_id: "Snowflake_Type",
        scheduled_event_id: "Snowflake_Type",
        payload: dict,
        reason: Absent[str] = MISSING,
    ) -> discord_typings.GuildScheduledEventData:
        """
        Modify a scheduled event for a guild.

        Args:
            guild_id: The guild to modify scheduled event from
            scheduled_event_id: The scheduled event to modify
            payload: The payload to modify the scheduled event with
            reason: The reason to be displayed in audit logs

        Returns:
            Scheduled Event or None

        """
        return await self.request(
            Route("PATCH", f"/guilds/{guild_id}/scheduled-events/{scheduled_event_id}"),
            payload=payload,
            reason=reason,
        )

    async def delete_scheduled_event(
        self,
        guild_id: "Snowflake_Type",
        scheduled_event_id: "Snowflake_Type",
        reason: Absent[str] = MISSING,
    ) -> None:
        """
        Delete a scheduled event for a guild.

        Args:
            guild_id: The guild to delete scheduled event from
            scheduled_event_id: The scheduled event to delete
            reason: The reason to be displayed in audit logs

        """
        return await self.request(
            Route("DELETE", f"/guilds/{guild_id}/scheduled-events/{scheduled_event_id}"),
            reason=reason,
        )

    async def get_scheduled_event_users(
        self,
        guild_id: "Snowflake_Type",
        scheduled_event_id: "Snowflake_Type",
        limit: int = 100,
        with_member: bool = False,
        before: "Snowflake_Type" = MISSING,
        after: "Snowflake_Type" = MISSING,
    ) -> List[discord_typings.GuildScheduledEventUserData]:
        """
        Get the users for a scheduled event.

        Args:
            guild_id: The guild to get scheduled event users from
            scheduled_event_id: The scheduled event to get users from
            limit: how many users to receive from the event
            with_member: include guild member data if it exists
            before: consider only users before given user id
            after: consider only users after given user id

        Returns:
            List of Scheduled Event Users or None

        """
        params = {"limit": limit, "with_member": with_member, "before": before, "after": after}
        return await self.request(
            Route("GET", f"/guilds/{guild_id}/scheduled-events/{scheduled_event_id}/users"),
            params=params,
        )



from typing import TYPE_CHECKING, cast

import discord_typings

from interactions.models.internal.protocols import CanRequest
from ..route import Route

__all__ = ("MessageRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type
    from interactions import UPLOADABLE_TYPE


class MessageRequests(CanRequest):
    async def create_message(
        self,
        payload: dict,
        channel_id: "Snowflake_Type",
        files: list["UPLOADABLE_TYPE"] | None = None,  # todo type payload
    ) -> discord_typings.MessageData:
        """
        Send a message to the specified channel.

        Args:
            payload: The message to send
            channel_id: id of the channel to send message in
            files: Any files to send with this message

        Returns:
            The resulting message object

        """
        result = await self.request(
            Route("POST", f"/channels/{int(channel_id)}/messages"), payload=payload, files=files
        )
        return cast(discord_typings.MessageData, result)

    async def delete_message(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", reason: str | None = None
    ) -> None:
        """
        Deletes a message from the specified channel.

        Args:
            channel_id: The id of the channel to delete the message from
            message_id: The id of the message to delete
            reason: The reason for this action

        """
        await self.request(
            Route("DELETE", f"/channels/{int(channel_id)}/messages/{int(message_id)}"),
            reason=reason,
        )

    async def bulk_delete_messages(
        self,
        channel_id: "Snowflake_Type",
        message_ids: list["Snowflake_Type"],
        reason: str | None = None,
    ) -> None:
        """
        Delete multiple messages in a single request.

        Args:
            channel_id: The id of the channel these messages are in
            message_ids: A list of message ids to delete
            reason: The reason for this action

        """
        payload = {"messages": [int(message_id) for message_id in message_ids]}

        await self.request(
            Route("POST", f"/channels/{int(channel_id)}/messages/bulk-delete"),
            payload=payload,
            reason=reason,
        )

    async def get_message(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type"
    ) -> discord_typings.MessageData:
        """
        Get a specific message in the channel. Returns a message object on success.

        Args:
            channel_id: the channel this message belongs to
            message_id: the id of the message

        Returns:
            message or None

        """
        result = await self.request(Route("GET", f"/channels/{int(channel_id)}/messages/{int(message_id)}"))
        return cast(discord_typings.MessageData, result)

    async def pin_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> None:
        """
        Pin a message to a channel.

        Args:
            channel_id: Channel to pin message to
            message_id: Message to pin

        """
        await self.request(Route("PUT", f"/channels/{int(channel_id)}/pins/{int(message_id)}"))

    async def unpin_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> None:
        """
        Unpin a message to a channel.

        Args:
            channel_id: Channel to unpin message to
            message_id: Message to unpin

        """
        await self.request(Route("DELETE", f"/channels/{int(channel_id)}/pins/{int(message_id)}"))

    async def edit_message(
        self,
        payload: dict,
        channel_id: "Snowflake_Type",
        message_id: "Snowflake_Type",
        files: list["UPLOADABLE_TYPE"] | None = None,
    ) -> discord_typings.MessageData:
        """
        Edit an existing message.

        Args:
            payload:
            channel_id: Channel of message to edit.
            message_id: Message to edit.
            files: Any files to send with this message

        Returns:
            Message object of edited message

        """
        result = await self.request(
            Route("PATCH", f"/channels/{int(channel_id)}/messages/{int(message_id)}"),
            payload=payload,
            files=files,
        )
        return cast(discord_typings.MessageData, result)

    async def crosspost_message(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type"
    ) -> discord_typings.MessageData:
        """
        Crosspost a message in a News Channel to following channels.

        Args:
            channel_id: Channel the message is in
            message_id: The id of the message to crosspost
        Returns:
            message object

        """
        result = await self.request(Route("POST", f"/channels/{int(channel_id)}/messages/{int(message_id)}/crosspost"))
        return cast(discord_typings.MessageData, result)



from .activity import Activity, ActivityAssets, ActivityParty, ActivitySecrets, ActivityTimestamps
from .app_perms import ApplicationCommandPermission
from .application import Application
from .asset import Asset
from .auto_mod import AutoModerationAction, AutoModRule
from .channel import (
    BaseChannel,
    ChannelHistory,
    DM,
    DMChannel,
    DMGroup,
    GuildCategory,
    GuildChannel,
    GuildForum,
    GuildForumPost,
    GuildNews,
    GuildNewsThread,
    GuildPrivateThread,
    GuildPublicThread,
    GuildStageVoice,
    GuildText,
    GuildVoice,
    InvitableMixin,
    MessageableMixin,
    PermissionOverwrite,
    process_permission_overwrites,
    ThreadableMixin,
    ThreadChannel,
    TYPE_ALL_CHANNEL,
    TYPE_CHANNEL_MAPPING,
    TYPE_DM_CHANNEL,
    TYPE_GUILD_CHANNEL,
    TYPE_MESSAGEABLE_CHANNEL,
    TYPE_THREAD_CHANNEL,
    TYPE_VOICE_CHANNEL,
    WebhookMixin,
)
from .color import (
    BrandColors,
    BrandColours,
    Color,
    COLOR_TYPES,
    Colour,
    FlatUIColors,
    FlatUIColours,
    MaterialColors,
    MaterialColours,
    process_color,
    process_colour,
    RoleColors,
    RoleColours,
)
from .components import (
    ActionRow,
    BaseComponent,
    BaseSelectMenu,
    Button,
    ChannelSelectMenu,
    get_components_ids,
    InteractiveComponent,
    MentionableSelectMenu,
    process_components,
    RoleSelectMenu,
    spread_to_rows,
    StringSelectMenu,
    StringSelectOption,
    TYPE_COMPONENT_MAPPING,
    UserSelectMenu,
)

from .embed import Embed, EmbedAttachment, EmbedAuthor, EmbedField, EmbedFooter, EmbedProvider, process_embeds
from .emoji import CustomEmoji, PartialEmoji, process_emoji, process_emoji_req_format
from .enums import (
    ActivityFlag,
    ActivityType,
    ApplicationFlags,
    AuditLogEventType,
    AutoArchiveDuration,
    ButtonStyle,
    ChannelFlags,
    ChannelType,
    CommandType,
    ComponentType,
    DefaultNotificationLevel,
    ExplicitContentFilterLevel,
    ForumLayoutType,
    IntegrationExpireBehaviour,
    Intents,
    InteractionPermissionTypes,
    InteractionType,
    InviteTargetType,
    MemberFlags,
    MentionType,
    MessageActivityType,
    MessageFlags,
    MessageType,
    MFALevel,
    NSFWLevel,
    OverwriteType,
    Permissions,
    PremiumTier,
    PremiumType,
    ScheduledEventPrivacyLevel,
    ScheduledEventStatus,
    ScheduledEventType,
    StagePrivacyLevel,
    Status,
    StickerFormatType,
    StickerTypes,
    SystemChannelFlags,
    TeamMembershipState,
    UserFlags,
    VerificationLevel,
    VideoQualityMode,
    WebSocketOPCode,
)
from .file import File, open_file, UPLOADABLE_TYPE
from .guild import (
    AuditLog,
    AuditLogChange,
    AuditLogEntry,
    AuditLogHistory,
    BaseGuild,
    Guild,
    GuildBan,
    GuildIntegration,
    GuildPreview,
    GuildTemplate,
    GuildWelcome,
    GuildWelcomeChannel,
    GuildWidget,
    GuildWidgetSettings,
)
from .invite import Invite
from .message import (
    AllowedMentions,
    Attachment,
    BaseMessage,
    ChannelMention,
    Message,
    MessageActivity,
    MessageInteraction,
    MessageReference,
    process_allowed_mentions,
    process_message_payload,
    process_message_reference,
)
from .modal import InputText, Modal, ParagraphText, ShortText, TextStyles
from .reaction import Reaction, ReactionUsers
from .role import Role
from .scheduled_event import ScheduledEvent
from .snowflake import (
    Snowflake,
    Snowflake_Type,
    SnowflakeObject,
    to_optional_snowflake,
    to_snowflake,
    to_snowflake_list,
)
from .stage_instance import StageInstance
from .sticker import Sticker, StickerItem, StickerPack
from .team import Team, TeamMember
from .thread import ThreadList, ThreadMember, ThreadTag
from .timestamp import Timestamp, TimestampStyles
from .user import BaseUser, Member, User, ClientUser
from .voice_state import VoiceRegion, VoiceState

__all__ = (
    "ActionRow",
    "Activity",
    "ActivityAssets",
    "ActivityFlag",
    "ActivityParty",
    "ActivitySecrets",
    "ActivityTimestamps",
    "ActivityType",
    "AllowedMentions",
    "Application",
    "ApplicationCommandPermission",
    "ApplicationFlags",
    "Asset",
    "Attachment",
    "AuditLog",
    "AuditLogChange",
    "AuditLogEntry",
    "AuditLogEventType",
    "AuditLogHistory",
    "AutoArchiveDuration",
    "AutoModerationAction",
    "AutoModRule",
    "BaseChannel",
    "BaseComponent",
    "BaseGuild",
    "BaseMessage",
    "BaseSelectMenu",
    "BaseUser",
    "BrandColors",
    "BrandColours",
    "Button",
    "ButtonStyle",
    "ChannelFlags",
    "ChannelHistory",
    "ChannelMention",
    "ChannelSelectMenu",
    "ChannelType",
    "ClientUser",
    "Color",
    "COLOR_TYPES",
    "Colour",
    "CommandType",
    "ComponentType",
    "CustomEmoji",
    "DefaultNotificationLevel",
    "DM",
    "DMChannel",
    "DMGroup",
    "Embed",
    "EmbedAttachment",
    "EmbedAuthor",
    "EmbedField",
    "EmbedFooter",
    "EmbedProvider",
    "ExplicitContentFilterLevel",
    "File",
    "FlatUIColors",
    "FlatUIColours",
    "ForumLayoutType",
    "get_components_ids",
    "Guild",
    "GuildBan",
    "GuildCategory",
    "GuildChannel",
    "GuildForum",
    "GuildForumPost",
    "GuildIntegration",
    "GuildNews",
    "GuildNewsThread",
    "GuildPreview",
    "GuildPrivateThread",
    "GuildPublicThread",
    "GuildStageVoice",
    "GuildTemplate",
    "GuildText",
    "GuildVoice",
    "GuildWelcome",
    "GuildWelcomeChannel",
    "GuildWidget",
    "GuildWidgetSettings",
    "InputText",
    "IntegrationExpireBehaviour",
    "Intents",
    "InteractionPermissionTypes",
    "InteractionType",
    "InteractiveComponent",
    "InvitableMixin",
    "Invite",
    "InviteTargetType",
    "MaterialColors",
    "MaterialColours",
    "Member",
    "MemberFlags",
    "MentionableSelectMenu",
    "MentionType",
    "Message",
    "MessageableMixin",
    "MessageActivity",
    "MessageActivityType",
    "MessageFlags",
    "MessageInteraction",
    "MessageReference",
    "MessageType",
    "MFALevel",
    "Modal",
    "NSFWLevel",
    "open_file",
    "OverwriteType",
    "ParagraphText",
    "PartialEmoji",
    "PermissionOverwrite",
    "Permissions",
    "PremiumTier",
    "PremiumType",
    "process_allowed_mentions",
    "process_color",
    "process_colour",
    "process_components",
    "process_embeds",
    "process_emoji",
    "process_emoji_req_format",
    "process_message_payload",
    "process_message_reference",
    "process_permission_overwrites",
    "Reaction",
    "ReactionUsers",
    "Role",
    "RoleColors",
    "RoleColours",
    "RoleSelectMenu",
    "ScheduledEvent",
    "ScheduledEventPrivacyLevel",
    "ScheduledEventStatus",
    "ScheduledEventType",
    "ShortText",
    "Snowflake",
    "Snowflake_Type",
    "SnowflakeObject",
    "spread_to_rows",
    "StageInstance",
    "StagePrivacyLevel",
    "Status",
    "Sticker",
    "StickerFormatType",
    "StickerItem",
    "StickerPack",
    "StickerTypes",
    "StringSelectMenu",
    "StringSelectOption",
    "SystemChannelFlags",
    "Team",
    "TeamMember",
    "TeamMembershipState",
    "TextStyles",
    "ThreadableMixin",
    "ThreadChannel",
    "ThreadList",
    "ThreadMember",
    "ThreadTag",
    "Timestamp",
    "TimestampStyles",
    "to_optional_snowflake",
    "to_snowflake",
    "to_snowflake_list",
    "TYPE_ALL_CHANNEL",
    "TYPE_CHANNEL_MAPPING",
    "TYPE_COMPONENT_MAPPING",
    "TYPE_DM_CHANNEL",
    "TYPE_GUILD_CHANNEL",
    "TYPE_MESSAGEABLE_CHANNEL",
    "TYPE_THREAD_CHANNEL",
    "TYPE_VOICE_CHANNEL",
    "UPLOADABLE_TYPE",
    "User",
    "UserFlags",
    "UserSelectMenu",
    "VerificationLevel",
    "VideoQualityMode",
    "VoiceRegion",
    "VoiceState",
    "Webhook",
    "WebhookMixin",
    "WebhookTypes",
    "WebSocketOPCode",
)
from .webhooks import Webhook, WebhookTypes



from asyncio import QueueEmpty
from collections import namedtuple
from typing import TYPE_CHECKING, List, Optional

import attrs

from interactions.client.const import MISSING
from interactions.models.discord.emoji import PartialEmoji
from interactions.models.discord.snowflake import to_snowflake
from interactions.models.misc.iterator import AsyncIterator
from .base import ClientObject

if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type
    from interactions.models import Message, TYPE_ALL_CHANNEL
    from interactions.models.discord.user import User

__all__ = ("ReactionUsers", "Reaction")


class ReactionUsers(AsyncIterator):
    """
    An async iterator for searching through a channel's history.

    Attributes:
        reaction: The reaction to search through
        limit: The maximum number of users to return (set to 0 for no limit)
        after: get users after this message ID

    """

    def __init__(self, reaction: "Reaction", limit: int = 50, after: Optional["Snowflake_Type"] = None) -> None:
        self.reaction: "Reaction" = reaction
        self.after: "Snowflake_Type" = after
        self._more = True
        super().__init__(limit)

    async def fetch(self) -> List["User"]:
        """
        Gets all the users who reacted to the message. Requests user data from discord API if not cached.

        Returns:
            A list of users who reacted to the message.

        """
        if self._more:
            expected = self.get_limit

            if self.after and not self.last:
                self.last = namedtuple("temp", "id")
                self.last.id = self.after

            users = await self.reaction._client.http.get_reactions(
                self.reaction._channel_id,
                self.reaction._message_id,
                self.reaction.emoji.req_format,
                limit=expected,
                after=self.last.id or MISSING,
            )
            if not users:
                raise QueueEmpty
            self._more = len(users) == expected
            return [self.reaction._client.cache.place_user_data(u) for u in users]
        raise QueueEmpty


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Reaction(ClientObject):
    count: int = attrs.field(
        repr=False,
    )
    """times this emoji has been used to react"""
    me: bool = attrs.field(repr=False, default=False)
    """whether the current user reacted using this emoji"""
    emoji: "PartialEmoji" = attrs.field(repr=False, converter=PartialEmoji.from_dict)
    """emoji information"""

    _channel_id: "Snowflake_Type" = attrs.field(repr=False, converter=to_snowflake)
    _message_id: "Snowflake_Type" = attrs.field(repr=False, converter=to_snowflake)

    def users(self, limit: int = 0, after: "Snowflake_Type" = None) -> ReactionUsers:
        """Users who reacted using this emoji."""
        return ReactionUsers(self, limit, after)

    @property
    def message(self) -> "Message":
        """The message this reaction is on."""
        return self._client.cache.get_message(self._channel_id, self._message_id)

    @property
    def channel(self) -> "TYPE_ALL_CHANNEL":
        """The channel this reaction is on."""
        return self._client.cache.get_channel(self._channel_id)

    async def remove(self) -> None:
        """Remove all this emoji's reactions from the message."""
        await self._client.http.clear_reaction(self._channel_id, self._message_id, self.emoji.req_format)



import enum
from typing import Optional


__all__ = (
    "bold",
    "italic",
    "underline",
    "strikethrough",
    "spoiler",
    "no_embed_link",
    "link_in_embed",
    "quote_line",
    "inline_code",
    "code_block",
    "ansi_block",
    "AnsiStyles",
    "AnsiColors",
    "AnsiBackgrounds",
    "styles",
    "colors",
    "bg_colors",
    "ansi_format",
    "ansi_escape",
    "ansi_styled",
)


def bold(text: str) -> str:
    """Formats text for discord message as bold"""
    return f"**{text}**"


def italic(text: str) -> str:
    """Formats text for discord message as italic"""
    return f"*{text}*"


def underline(text: str) -> str:
    """Formats text for discord message as underlined"""
    return f"__{text}__"


def strikethrough(text: str) -> str:
    """Formats text for discord message as strikethrough"""
    return f"~~{text}~~"


def spoiler(text: str) -> str:
    """Formats text for discord message as spoiler"""
    return f"||{text}||"


def no_embed_link(url: str) -> str:
    """Makes link in discord message display without embedded website preview"""
    return f"<{url}>"


def link_in_embed(text: str, url: str) -> str:
    """Makes a clickable link inside Embed object"""
    return f"[{text}]({url})"


def quote_line(line: str) -> str:
    """Formats a text line for discord message as quote"""
    return f"> {line}"


def inline_code(text: str) -> str:
    """Formats text for discord message as inline code"""
    return f"`{text}`"


def code_block(text: str, language: Optional[str]) -> str:
    """Formats text for discord message as code block"""
    return f"```{language or ''}\n" f"{text}" f"```"


def ansi_block(text: str) -> str:
    """Formats text for discord message as code block that allows for arbitrary coloring and formatting"""
    return code_block(text, "ansi")


class AnsiStyles(enum.IntEnum):
    NORMAL = 0
    BOLD = 1
    UNDERLINE = 4


class AnsiColors(enum.IntEnum):
    GRAY = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    PINK = 35
    CYAN = 36
    WHITE = 37


class AnsiBackgrounds(enum.IntEnum):
    SOME_VERY_DARK_BLUE = 40
    ORANGE = 41
    GRAY = 42
    LIGHT_GRAY = 43
    EVEN_LIGHTER_GRAY = 44
    INDIGO = 45
    AGAIN_LIGHTER_GRAY = 46
    WHITE = 47


# Just short aliases
styles = AnsiStyles
colors = AnsiColors
bg_colors = AnsiBackgrounds


def ansi_format(
    style: Optional[AnsiStyles] = None,
    color: Optional[AnsiColors] = None,
    background: Optional[AnsiBackgrounds] = None,
) -> str:
    """Gives format prefix for ansi code block with selected styles"""
    text_style = ";".join(str(_style.value) for _style in (style, color, background) if _style)
    return f"[{text_style}m"


ansi_escape = "[0m"


def ansi_styled(
    text: str,
    style: Optional[AnsiStyles] = None,
    color: Optional[AnsiColors] = None,
    background: Optional[AnsiBackgrounds] = None,
) -> str:
    """Formats text for ansi code block with selected styles"""
    return f"{ansi_format(style, color, background)}{text}{ansi_escape}"



from asyncio import sleep, Task, create_task
from typing import TYPE_CHECKING

from interactions.client.const import Absent, MISSING

if TYPE_CHECKING:
    from interactions import MessageableMixin

__all__ = ("Typing",)


class Typing:
    """
    A context manager to send a typing state to a given channel as long as long as the wrapped operation takes.

    Args:
        channel: The channel to send the typing state to
    """

    def __init__(self, channel: "MessageableMixin") -> None:
        self._task: Absent[Task] = MISSING
        self.channel: "MessageableMixin" = channel
        self._stop: bool = False

    async def _typing_task(self) -> None:
        while not self._stop:
            await self.channel.trigger_typing()
            await sleep(5)

    async def __aenter__(self) -> None:
        self.task = create_task(self._typing_task())

    async def __aexit__(self, *_) -> None:
        self._stop = True
        self.task.cancel()



import asyncio

import aiomonitor as aiomonitor

import interactions
from interactions import Extension

__all__ = ("Console",)


class Console(Extension):
    """
    Extension that starts the bot with the aiomonitor console active - notably giving you REPL for the bot

    To access the console, you need to connect to the port specified in the constructor, by default 501.
    On linux, you can do this with `nc localhost 501`, on windows you can use `telnet localhost 501`.

    For both platforms you can also use "python -m aiomonitor.cli -p 501" as a replacement for the above commands.

    Args:
        port: The port to start the aiomonitor on
        console_port: The port to start the aiomonitor console on
        **kwargs: The locals to make available in the console, by default this includes `client`, `bot` and `interactions`

    """

    def __init__(self, client: interactions.Client, port: int = 501, console_port: int = 502, **kwargs) -> None:
        self.client.astart = self.async_start_bot
        self.port = port  # 501 was chosen as windows throws a massive fit if you try to use the default port
        self.console_port = console_port
        self.locals = kwargs

    async def async_start_bot(self, token: str | None = None) -> None:
        """Starts the bot with the console active"""
        old_start = interactions.Client.astart

        _locals = self.locals | {
            "client": self.client,
            "bot": self.client,
            "interactions": interactions,
        }

        with aiomonitor.start_monitor(
            loop=asyncio.get_event_loop(), port=self.port, console_port=self.console_port, locals=_locals
        ) as monitor:
            self.client.logger.info(f"Started aiomonitor on {monitor._host}:{monitor._port}")

            await old_start(self.client, token)



from typing import TYPE_CHECKING

from interactions.models.discord.app_perms import ApplicationCommandPermission, CommandPermissions
from interactions.models.discord.snowflake import to_snowflake
from ._template import EventMixinTemplate, Processor
from ... import events

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("IntegrationEvents",)


class IntegrationEvents(EventMixinTemplate):
    @Processor.define()
    async def _raw_application_command_permissions_update(self, event: "RawGatewayEvent") -> None:
        perms = [ApplicationCommandPermission.from_dict(perm, self) for perm in event.data["permissions"]]
        guild_id = to_snowflake(event.data["guild_id"])
        command_id = to_snowflake(event.data["id"])

        if guild := self.get_guild(guild_id):
            if guild.permissions:
                if command_id not in guild.command_permissions:
                    guild.command_permissions[command_id] = CommandPermissions(
                        client=self, command_id=command_id, guild=guild
                    )

                command_permissions = guild.command_permissions[command_id]
                command_permissions.update_permissions(*perms)

        self.dispatch(events.ApplicationCommandPermissionsUpdate(guild, perms))



from .slash import (
    slash_attachment_option,
    slash_bool_option,
    slash_channel_option,
    slash_float_option,
    slash_int_option,
    slash_mentionable_option,
    slash_role_option,
    slash_str_option,
    slash_user_option,
)

__all__ = (
    "slash_attachment_option",
    "slash_bool_option",
    "slash_channel_option",
    "slash_float_option",
    "slash_int_option",
    "slash_mentionable_option",
    "slash_role_option",
    "slash_str_option",
    "slash_user_option",
)



from typing import TYPE_CHECKING, Any, List

import discord_typings

from ..route import Route

__all__ = ("UserRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type


class UserRequests:
    request: Any

    async def get_current_user(self) -> discord_typings.UserData:
        """
        Shortcut to get requester's user.

        Returns:
            The user object.

        """
        return await self.get_user("@me")

    async def get_user(self, user_id: "Snowflake_Type") -> discord_typings.UserData:
        """
        Get a user object for a given user ID.

        Args:
            user_id: The user to get.

        Returns:
            The user object.

        """
        return await self.request(Route("GET", f"/users/{user_id}"))

    async def modify_client_user(self, payload: dict) -> discord_typings.UserData:
        """
        Modify the user account settings.

        Args:
            payload: The data to send.

        """
        return await self.request(Route("PATCH", "/users/@me"), payload=payload)

    async def get_user_guilds(self) -> List[discord_typings.GuildData]:
        """
        Returns a list of partial guild objects the current user is a member of.

        Requires the guilds OAuth2 scope.

        """
        return await self.request(Route("GET", "/users/@me/guilds"))

    async def leave_guild(self, guild_id: "Snowflake_Type") -> None:
        """
        Leave a guild. Returns a 204 empty response on success.

        Args:
            guild_id: The guild to leave from.

        """
        return await self.request(Route("DELETE", f"/users/@me/guilds/{guild_id}"))

    async def create_dm(self, recipient_id: "Snowflake_Type") -> discord_typings.DMChannelData:
        """
        Create a new DM channel with a user. Returns a DM channel object.

        Args:
            recipient_id: The recipient to open a DM channel with.

        """
        return await self.request(Route("POST", "/users/@me/channels"), payload={"recipient_id": recipient_id})

    async def create_group_dm(self, payload: dict) -> discord_typings.GroupDMChannelData:
        """
        Create a new group DM channel with multiple users.

        Args:
            payload: The data to send.

        """
        return await self.request(Route("POST", "/users/@me/channels"), payload=payload)

    async def get_user_connections(self) -> list:
        """
        Returns a list of connection objects.

        Requires the connections OAuth2 scope.

        """
        return await self.request(Route("GET", "/users/@me/connections"))

    async def group_dm_add_recipient(
        self,
        channel_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        access_token: str,
        nick: str = None,
    ) -> None:
        """
        Adds a recipient to a Group DM using their access token.

        Args:
            channel_id: The ID of the group dm
            user_id: The ID of the user to add
            access_token: Access token of a user that has granted your app the gdm.join scope
            nick: Nickname of the user being added

        """
        return await self.request(
            Route("PUT", f"/channels/{channel_id}/recipients/{user_id}"),
            payload={"access_token": access_token, "nick": nick},
        )

    async def group_dm_remove_recipient(self, channel_id: "Snowflake_Type", user_id: "Snowflake_Type") -> None:
        """
        Remove a recipient from the group dm.

        Args:
            channel_id: The ID of the group dm
            user_id: The ID of the user to remove

        """
        return await self.request(Route("DELETE", f"/channels/{channel_id}/recipients/{user_id}"))

    async def modify_current_user_nick(self, guild_id: "Snowflake_Type", nickname: str = None) -> None:
        """
        Modifies the nickname of the current user in a guild.

        Args:
            guild_id: The ID of the guild
            nickname: The new nickname to use

        """
        return await self.request(Route("PATCH", f"/guilds/{guild_id}/members/@me/nick"), payload={"nick": nickname})



from .triggers import BaseTrigger, IntervalTrigger, DateTrigger, TimeTrigger, OrTrigger
from .task import Task

__all__ = ("BaseTrigger", "IntervalTrigger", "DateTrigger", "TimeTrigger", "OrTrigger", "Task")



from typing import Callable, Any, Coroutine

from interactions.client.client import Client
from interactions.models.discord.message import Message

__all__ = ("when_mentioned", "when_mentioned_or")


async def when_mentioned(bot: Client, _) -> list[str]:
    """
    Returns a list of the bot's mentions.

    Returns:
        A list of the bot's possible mentions.
    """
    return [f"<@{bot.user.id}> ", f"<@!{bot.user.id}> "]  # type: ignore


def when_mentioned_or(
    *prefixes: str,
) -> Callable[[Client, Message], Coroutine[Any, Any, list[str]]]:
    """
    Returns a list of the bot's mentions plus whatever prefixes are provided.

    This is intended to be used with initializing prefixed commands. If you wish to use
    it in your own function, you will need to do something similar to
    `await when_mentioned_or(*prefixes)(bot, msg)`.

    Args:
        prefixes: Prefixes to include alongside mentions.

    Returns:
        A list of the bot's mentions plus whatever prefixes are provided.
    """

    async def _new_mention(bot: Client, _) -> list[str]:
        return (await when_mentioned(bot, _)) + list(prefixes)

    return _new_mention



from typing import TYPE_CHECKING, Optional, Union, Dict, Any

import attrs

from interactions.client.const import MISSING, Absent
from interactions.client.utils.attr_converters import optional as optional_c
from interactions.client.utils.attr_converters import timestamp_converter
from interactions.models.discord.application import Application
from interactions.models.discord.enums import InviteTargetType
from interactions.models.discord.guild import GuildPreview
from interactions.models.discord.snowflake import to_snowflake
from interactions.models.discord.stage_instance import StageInstance
from interactions.models.discord.timestamp import Timestamp
from .base import ClientObject

if TYPE_CHECKING:
    from interactions.client import Client
    from interactions.models import TYPE_GUILD_CHANNEL
    from interactions.models.discord.user import User
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("Invite",)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Invite(ClientObject):
    code: str = attrs.field(repr=True)
    """the invite code (unique ID)"""

    # metadata
    uses: int = attrs.field(default=0, repr=True)
    """the guild this invite is for"""
    max_uses: int = attrs.field(repr=False, default=0)
    """max number of times this invite can be used"""
    max_age: int = attrs.field(repr=False, default=0)
    """duration (in seconds) after which the invite expires"""
    created_at: Timestamp = attrs.field(default=MISSING, converter=optional_c(timestamp_converter), repr=True)
    """when this invite was created"""
    temporary: bool = attrs.field(default=False, repr=True)
    """whether this invite only grants temporary membership"""

    # target data
    target_type: Optional[Union[InviteTargetType, int]] = attrs.field(
        default=None, converter=optional_c(InviteTargetType), repr=True
    )
    """the type of target for this voice channel invite"""
    approximate_presence_count: Optional[int] = attrs.field(repr=False, default=MISSING)
    """approximate count of online members, returned from the `GET /invites/<code>` endpoint when `with_counts` is `True`"""
    approximate_member_count: Optional[int] = attrs.field(repr=False, default=MISSING)
    """approximate count of total members, returned from the `GET /invites/<code>` endpoint when `with_counts` is `True`"""
    scheduled_event: Optional["Snowflake_Type"] = attrs.field(
        default=None, converter=optional_c(to_snowflake), repr=True
    )
    """guild scheduled event data, only included if `guild_scheduled_event_id` contains a valid guild scheduled event id"""
    expires_at: Optional[Timestamp] = attrs.field(default=None, converter=optional_c(timestamp_converter), repr=True)
    """the expiration date of this invite, returned from the `GET /invites/<code>` endpoint when `with_expiration` is `True`"""
    stage_instance: Optional[StageInstance] = attrs.field(repr=False, default=None)
    """stage instance data if there is a public Stage instance in the Stage channel this invite is for (deprecated)"""
    target_application: Optional[dict] = attrs.field(repr=False, default=None)
    """the embedded application to open for this voice channel embedded application invite"""
    guild_preview: Optional[GuildPreview] = attrs.field(repr=False, default=MISSING)
    """the guild this invite is for"""

    # internal for props
    _channel_id: "Snowflake_Type" = attrs.field(converter=to_snowflake, repr=True)
    _inviter_id: Optional["Snowflake_Type"] = attrs.field(default=None, converter=optional_c(to_snowflake), repr=True)
    _target_user_id: Optional["Snowflake_Type"] = attrs.field(
        repr=False, default=None, converter=optional_c(to_snowflake)
    )

    @property
    def channel(self) -> "TYPE_GUILD_CHANNEL":
        """The channel the invite is for."""
        return self._client.cache.get_channel(self._channel_id)

    @property
    def inviter(self) -> Optional["User"]:
        """The user that created the invite or None."""
        return self._client.cache.get_user(self._inviter_id) if self._inviter_id else None

    @property
    def target_user(self) -> Optional["User"]:
        """The user whose stream to display for this voice channel stream invite or None."""
        return self._client.cache.get_user(self._target_user_id) if self._target_user_id else None

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        if "stage_instance" in data:
            data["stage_instance"] = StageInstance.from_dict(data, client)

        if "target_application" in data:
            data["target_application"] = Application.from_dict(data, client)

        if "target_event_id" in data:
            data["scheduled_event"] = data["target_event_id"]

        if channel := data.pop("channel", None):
            # invite metadata does not contain enough info to create a channel object
            data["channel_id"] = channel["id"]

        if guild := data.pop("guild", None):
            data["guild_preview"] = GuildPreview.from_dict(guild, client)

        if inviter := data.pop("inviter", None):
            inviter = client.cache.place_user_data(inviter)
            data["inviter_id"] = inviter.id

        return data

    def __str__(self) -> str:
        return self.link

    @property
    def link(self) -> str:
        """The invite link."""
        if self.scheduled_event:
            return f"https://discord.gg/{self.code}?event={self.scheduled_event}"
        return f"https://discord.gg/{self.code}"

    async def delete(self, reason: Absent[str] = MISSING) -> None:
        """
        Delete this invite.

        !!! note
            You must have the `manage_channels` permission on the channel this invite belongs to.

        !!! note
            With `manage_guild` permission, you can delete any invite across the guild.

        Args:
            reason: The reason for the deletion of invite.

        """
        await self._client.http.delete_invite(self.code, reason=reason)



import copy
from typing import Any, TypeVar, Coroutine, Callable

__all__ = ("CallbackObject",)

T = TypeVar("T", bound="CallbackObject")


class CallbackObject:
    _binding: Any
    callback: Callable[..., Coroutine[Any, Any, Any]]

    def __new__(cls, *args, **kwargs) -> T:
        self = super().__new__(cls)

        self._binding = None

        return self

    @property
    def has_binding(self) -> bool:
        return self._binding is not None

    def copy_with_binding(self, binding: Any = None) -> T:
        obj = copy.copy(self)
        obj._binding = binding
        return obj

    async def __call__(self, *args, **kwargs) -> Any:
        if self._binding:
            return await self.callback(self._binding, *args, **kwargs)
        return await self.callback(*args, **kwargs)

    async def call_with_binding(self, callback: Callable[..., Coroutine[Any, Any, Any]], *args, **kwargs) -> Any:
        """
        Call a given method using this objects _binding.

        Args:
            callback: The callback to call.
        """
        if self._binding:
            return await callback(self._binding, *args, **kwargs)
        return await callback(*args, **kwargs)



import typing

from interactions.models.discord.snowflake import Snowflake

if typing.TYPE_CHECKING:
    import interactions

__all__ = ("ModalMixin",)


class ModalMixin:
    client: "interactions.Client"
    """The client that created this context."""
    responded: bool
    """Whether this context has been responded to."""
    id: Snowflake
    """The interaction ID."""
    token: str
    """The interaction token."""

    async def send_modal(self, modal: "interactions.Modal") -> "dict | interactions.Modal":
        """Send a modal to the user."""
        if self.responded:
            raise RuntimeError("Cannot send modal after responding")
        payload = modal if isinstance(modal, dict) else modal.to_dict()

        await self.client.http.post_initial_response(payload, self.id, self.token)

        self.responded = True
        return modal



from typing import TYPE_CHECKING, cast

import discord_typings

from interactions.client.const import GLOBAL_SCOPE
from interactions.models.internal.protocols import CanRequest
from ..route import Route

__all__ = ("InteractionRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type
    from interactions import UPLOADABLE_TYPE


class InteractionRequests(CanRequest):
    async def delete_application_command(
        self,
        application_id: "Snowflake_Type",
        guild_id: "Snowflake_Type",
        command_id: "Snowflake_Type",
    ) -> None:
        """
        Delete an existing application command for this application.

        Args:
            application_id: the what application to delete for
            guild_id: specify a guild to delete commands from
            command_id: the command to delete

        """
        if guild_id == GLOBAL_SCOPE:
            await self.request(Route("DELETE", f"/applications/{int(application_id)}/commands/{int(command_id)}"))
        else:
            await self.request(
                Route(
                    "DELETE",
                    f"/applications/{int(application_id)}/guilds/{int(guild_id)}/commands/{int(command_id)}",
                )
            )

    async def get_application_commands(
        self,
        application_id: "Snowflake_Type",
        guild_id: "Snowflake_Type",
        with_localisations: bool = True,
    ) -> list[discord_typings.ApplicationCommandData]:
        """
        Get all application commands for this application from discord.

        Args:
            application_id: the what application to query
            guild_id: specify a guild to get commands from
            with_localisations: whether to include all localisations in the response

        Returns:
            Application command data

        """
        if guild_id == GLOBAL_SCOPE:
            return await self.request(
                Route("GET", f"/applications/{application_id}/commands"),
                params={"with_localizations": int(with_localisations)},
            )
        return await self.request(
            Route("GET", f"/applications/{application_id}/guilds/{guild_id}/commands"),
            params={"with_localizations": int(with_localisations)},
        )

    async def overwrite_application_commands(
        self, app_id: "Snowflake_Type", data: list[dict], guild_id: "Snowflake_Type"
    ) -> list[discord_typings.ApplicationCommandData]:  # todo type "data"
        """
        Take a list of commands and overwrite the existing command list within the given scope

        Args:
            app_id: The application ID of this bot
            guild_id: The ID of the guild this command is for, if this is a guild command
            data: List of your interaction data

        """
        if guild_id == GLOBAL_SCOPE:
            result = await self.request(Route("PUT", f"/applications/{app_id}/commands"), payload=data)
        else:
            result = await self.request(
                Route("PUT", f"/applications/{app_id}/guilds/{int(guild_id)}/commands"),
                payload=data,
            )
        return cast(list[discord_typings.ApplicationCommandData], result)

    async def create_application_command(
        self, app_id: "Snowflake_Type", command: dict, guild_id: "Snowflake_Type"
    ) -> discord_typings.ApplicationCommandData:
        """
        Add a given command to scope.

        Args:
            app_id: The application ID of this bot
            command: A dictionary representing a command to be created
            guild_id: The ID of the guild this command is for, if this is a guild command

        Returns:
            An application command object
        """
        if guild_id == GLOBAL_SCOPE:
            result = await self.request(Route("POST", f"/applications/{app_id}/commands"), payload=command)
        else:
            result = await self.request(
                Route("POST", f"/applications/{app_id}/guilds/{int(guild_id)}/commands"),
                payload=command,
            )
        return cast(discord_typings.ApplicationCommandData, result)

    async def post_initial_response(
        self,
        payload: dict,
        interaction_id: str,
        token: str,
        files: list["UPLOADABLE_TYPE"] | None = None,
    ) -> None:
        """
        Post an initial response to an interaction.

        Args:
            payload: the payload to send
            interaction_id: the id of the interaction
            token: the token of the interaction
            files: The files to send in this message

        """
        return await self.request(
            Route("POST", f"/interactions/{interaction_id}/{token}/callback"),
            payload=payload,
            files=files,
        )

    async def post_followup(
        self,
        payload: dict,
        application_id: "Snowflake_Type",
        token: str,
        files: list["UPLOADABLE_TYPE"] | None = None,
    ) -> None:
        """
        Send a followup to an interaction.

        Args:
            payload: the payload to send
            application_id: the id of the application
            token: the token of the interaction
            files: The files to send with this interaction

        """
        return await self.request(
            Route("POST", f"/webhooks/{int(application_id)}/{token}"), payload=payload, files=files
        )

    async def edit_interaction_message(
        self,
        payload: dict,
        application_id: "Snowflake_Type",
        token: str,
        message_id: "str|Snowflake_Type" = "@original",
        files: list["UPLOADABLE_TYPE"] | None = None,
    ) -> discord_typings.MessageData:
        """
        Edits an existing interaction message.

        Args:
            payload: The payload to send.
            application_id: The id of the application.
            token: The token of the interaction.
            message_id: The target message to edit. Defaults to @original which represents the initial response message.
            files: The files to send with this interaction

        Returns:
            The edited message data.

        """
        result = await self.request(
            Route("PATCH", f"/webhooks/{int(application_id)}/{token}/messages/{message_id}"),
            payload=payload,
            files=files,
        )
        return cast(discord_typings.MessageData, result)

    async def delete_interaction_message(
        self,
        application_id: "Snowflake_Type",
        token: str,
        message_id: "str | Snowflake_Type" = "@original",
    ) -> None:
        """
        Deletes an existing interaction message.

        Args:
            application_id: The id of the application.
            token: The token of the interaction.
            message_id: The target message to delete. Defaults to @original which represents the initial response message.

        """
        return await self.request(Route("DELETE", f"/webhooks/{int(application_id)}/{token}/messages/{message_id}"))

    async def get_interaction_message(
        self, application_id: "Snowflake_Type", token: str, message_id: str = "@original"
    ) -> discord_typings.MessageData:
        """
        Gets an existing interaction message.

        Args:
            application_id: The id of the application.
            token: The token of the interaction.
            message_id: The target message to get. Defaults to @original which represents the initial response message.

        Returns:
            The message data.

        """
        result = await self.request(Route("GET", f"/webhooks/{int(application_id)}/{token}/messages/{message_id}"))
        return cast(discord_typings.MessageData, result)

    async def edit_application_command_permissions(
        self,
        application_id: "Snowflake_Type",
        scope: "Snowflake_Type",
        command_id: "Snowflake_Type",
        permissions: list[dict],  # todo better typing
    ) -> discord_typings.ApplicationCommandPermissionsData:
        """
        Edits command permissions for a specific command.

        Args:
            application_id: the id of the application
            scope: The scope this command is in
            command_id: The command id to edit
            permissions: The permissions to set to this command

        Returns:
            Guild Application Command Permissions

        """
        result = await self.request(
            Route(
                "PUT",
                f"/applications/{int(application_id)}/guilds/{int(scope)}/commands/{int(command_id)}/permissions",
            ),
            payload=permissions,
        )
        return cast(discord_typings.ApplicationCommandPermissionsData, result)

    async def get_application_command_permissions(
        self,
        application_id: "Snowflake_Type",
        scope: "Snowflake_Type",
        command_id: "Snowflake_Type",
    ) -> list[discord_typings.ApplicationCommandPermissionsData]:
        """
        Get permission data for a command.

        Args:
            application_id: the id of the application
            scope: The scope this command is in
            command_id: The command id to edit

        Returns:
            guild application command permissions

        """
        result = await self.request(
            Route(
                "GET",
                f"/applications/{int(application_id)}/guilds/{int(scope)}/commands/{int(command_id)}/permissions",
            )
        )
        return cast(list[discord_typings.ApplicationCommandPermissionsData], result)

    async def batch_get_application_command_permissions(
        self, application_id: "Snowflake_Type", scope: "Snowflake_Type"
    ) -> list[discord_typings.GuildApplicationCommandPermissionData]:
        """
        Get permission data for all commands in a scope.

        Args:
            application_id: the id of the application
            scope: The scope this command is in

        Returns:
            list of guild application command permissions

        """
        result = await self.request(
            Route(
                "GET",
                f"/applications/{int(application_id)}/guilds/{int(scope)}/commands/permissions",
            )
        )
        return cast(list[discord_typings.GuildApplicationCommandPermissionData], result)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import attrs

from interactions.client.const import MISSING, Absent
from interactions.client.errors import EventLocationNotProvided
from interactions.client.utils import to_image_data
from interactions.client.utils.attr_converters import optional
from interactions.client.utils.attr_converters import timestamp_converter
from interactions.models.discord.asset import Asset
from interactions.models.discord.file import UPLOADABLE_TYPE
from interactions.models.discord.snowflake import Snowflake_Type, to_snowflake
from interactions.models.discord.timestamp import Timestamp
from .base import DiscordObject
from .enums import ScheduledEventPrivacyLevel, ScheduledEventType, ScheduledEventStatus

if TYPE_CHECKING:
    from interactions.client import Client
    from interactions.models.discord.channel import GuildStageVoice, GuildVoice
    from interactions.models.discord.guild import Guild
    from interactions.models.discord.user import Member
    from interactions.models.discord.user import User

__all__ = ("ScheduledEvent",)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ScheduledEvent(DiscordObject):
    name: str = attrs.field(repr=True)
    description: str = attrs.field(repr=False, default=MISSING)
    entity_type: Union[ScheduledEventType, int] = attrs.field(repr=False, converter=ScheduledEventType)
    """The type of the scheduled event"""
    start_time: Timestamp = attrs.field(repr=False, converter=timestamp_converter)
    """A Timestamp object representing the scheduled start time of the event """
    end_time: Optional[Timestamp] = attrs.field(repr=False, default=None, converter=optional(timestamp_converter))
    """Optional Timstamp object representing the scheduled end time, required if entity_type is EXTERNAL"""
    privacy_level: Union[ScheduledEventPrivacyLevel, int] = attrs.field(
        repr=False, converter=ScheduledEventPrivacyLevel
    )
    """
    Privacy level of the scheduled event

    ??? note
        Discord only has `GUILD_ONLY` at the momment.
    """
    status: Union[ScheduledEventStatus, int] = attrs.field(repr=False, converter=ScheduledEventStatus)
    """Current status of the scheduled event"""
    entity_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=MISSING, converter=optional(to_snowflake))
    """The id of an entity associated with a guild scheduled event"""
    entity_metadata: Optional[Dict[str, Any]] = attrs.field(repr=False, default=MISSING)  # TODO make this
    """The metadata associated with the entity_type"""
    user_count: int = attrs.field(repr=False, default=MISSING)
    """Amount of users subscribed to the scheduled event"""
    cover: Asset | None = attrs.field(repr=False, default=None)
    """The cover image of this event"""

    _guild_id: "Snowflake_Type" = attrs.field(repr=False, converter=to_snowflake)
    _creator: Optional["User"] = attrs.field(repr=False, default=MISSING)
    _creator_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=MISSING, converter=optional(to_snowflake))
    _channel_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=optional(to_snowflake))

    @property
    async def creator(self) -> Optional["User"]:
        """
        Returns the user who created this event.

        !!! note
            Events made before October 25th, 2021 will not have a creator.

        """
        return await self._client.cache.fetch_user(self._creator_id) if self._creator_id else None

    @property
    def guild(self) -> "Guild":
        return self._client.cache.get_guild(self._guild_id)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        if data.get("creator"):
            data["creator"] = client.cache.place_user_data(data["creator"])

        if data.get("channel_id"):
            data["channel"] = client.cache.get_channel(data["channel_id"])

        data["start_time"] = data.get("scheduled_start_time")

        if end_time := data.get("scheduled_end_time"):
            data["end_time"] = end_time
        else:
            data["end_time"] = None

        if image := data.get("image"):
            data["cover"] = Asset.from_path_hash(client, f"guild-events/{data['id']}/{{}}", image)

        data = super()._process_dict(data, client)
        return data

    @property
    def location(self) -> Optional[str]:
        """Returns the external locatian of this event."""
        if self.entity_type == ScheduledEventType.EXTERNAL:
            return self.entity_metadata["location"]
        return None

    async def fetch_channel(self) -> Optional[Union["GuildVoice", "GuildStageVoice"]]:
        """Returns the channel this event is scheduled in if it is scheduled in a channel."""
        if self._channel_id:
            return await self._client.cache.fetch_channel(self._channel_id)
        return None

    def get_channel(self) -> Optional[Union["GuildVoice", "GuildStageVoice"]]:
        """Returns the channel this event is scheduled in if it is scheduled in a channel."""
        if self._channel_id:
            return self._client.cache.get_channel(self._channel_id)
        return None

    async def fetch_event_users(
        self,
        limit: Optional[int] = 100,
        with_member_data: bool = False,
        before: Absent[Optional["Snowflake_Type"]] = MISSING,
        after: Absent[Optional["Snowflake_Type"]] = MISSING,
    ) -> List[Union["Member", "User"]]:
        """
        Fetch event users.

        Args:
            limit: Discord defualts to 100
            with_member_data: Whether to include guild member data
            before: Snowflake of a user to get before
            after: Snowflake of a user to get after

        !!! note
            This method is paginated

        """
        event_users = await self._client.http.get_scheduled_event_users(
            self._guild_id, self.id, limit, with_member_data, before, after
        )
        participants = []
        for u in event_users:
            if member := u.get("member"):
                u["member"]["user"] = u["user"]
                participants.append(self._client.cache.place_member_data(self._guild_id, member))
            else:
                participants.append(self._client.cache.place_user_data(u["user"]))

        return participants

    async def delete(self, reason: Absent[str] = MISSING) -> None:
        """
        Deletes this event.

        Args:
            reason: The reason for deleting this event

        """
        await self._client.http.delete_scheduled_event(self._guild_id, self.id, reason)

    async def edit(
        self,
        *,
        name: Absent[str] = MISSING,
        start_time: Absent["Timestamp"] = MISSING,
        end_time: Absent["Timestamp"] = MISSING,
        status: Absent[ScheduledEventStatus] = MISSING,
        description: Absent[str] = MISSING,
        channel_id: Absent[Optional["Snowflake_Type"]] = MISSING,
        event_type: Absent[ScheduledEventType] = MISSING,
        external_location: Absent[Optional[str]] = MISSING,
        entity_metadata: Absent[dict] = MISSING,
        privacy_level: Absent[ScheduledEventPrivacyLevel] = MISSING,
        cover_image: Absent[UPLOADABLE_TYPE] = MISSING,
        reason: Absent[str] = MISSING,
    ) -> None:
        """
        Edits this event.

        Args:
            name: The name of the event
            description: The description of the event
            channel_id: The channel id of the event
            event_type: The type of the event
            start_time: The scheduled start time of the event
            end_time: The scheduled end time of the event
            status: The status of the event
            external_location: The location of the event (1-100 characters)
            entity_metadata: The metadata of the event
            privacy_level: The privacy level of the event
            cover_image: the cover image of the scheduled event
            reason: The reason for editing the event

        !!! note
            If updating event_type to EXTERNAL:
                `channel_id` is required and must be set to null

                `external_location` or `entity_metadata` with a location field must be provided

                `end_time` must be provided

        """
        if external_location is not MISSING:
            entity_metadata = {"location": external_location}

        if event_type == ScheduledEventType.EXTERNAL:
            channel_id = None
            if external_location == MISSING:
                raise EventLocationNotProvided("Location is required for external events")

        payload = {
            "name": name,
            "description": description,
            "channel_id": channel_id,
            "entity_type": event_type,
            "scheduled_start_time": start_time.isoformat() if start_time else MISSING,
            "scheduled_end_time": end_time.isoformat() if end_time else MISSING,
            "status": status,
            "entity_metadata": entity_metadata,
            "privacy_level": privacy_level,
            "image": to_image_data(cover_image) if cover_image else MISSING,
        }
        await self._client.http.modify_scheduled_event(self._guild_id, self.id, payload, reason)



"""
Constants used throughout interactions.py.

Attributes:
    __version__ str: The version of the library.
    __repo_url__ str: The URL of the repository.
    __py_version__ str: The python version in use.
    logger_name str: The name of interactions.py's default logger. Invalid if a custom logger is passed to `Client` to replace the default logger.
    logger logging.Logger: The logger used throughout interactions.py. If a custom logger is passed to `Client`, this obj is replaced with the new logger.
    kwarg_spam bool: Should ``unused kwargs`` be logged.

    ACTION_ROW_MAX_ITEMS int: The maximum number of items in an action row.
    SELECTS_MAX_OPTIONS int: The maximum number of options a select may have.
    SELECT_MAX_NAME_LENGTH int: The max length of a select's name.

    CONTEXT_MENU_NAME_LENGTH int: The max length of a context menu's name.
    SLASH_CMD_NAME_LENGTH int: The max length of a slash command's name.
    SLASH_CMD_MAX_DESC_LENGTH int: The maximum length of a slash command's description.
    SLASH_CMD_MAX_OPTIONS int: The maximum number of options a slash command may have.
    SLASH_OPTION_NAME_LENGTH int: The maximum length of a slash option's name.

    EMBED_MAX_NAME_LENGTH int: The maximum length for an embed title
    EMBED_MAX_DESC_LENGTH int: The maximum length for an embed description
    EMBED_MAX_FIELDS int: The maximum number of fields for an embed
    EMBED_TOTAL_MAX int: The total combined number of characters for an embed
    PREMIUM_GUILD_LIMITS dict: Limits granted per premium level of a guild

    GLOBAL_SCOPE _sentinel: A sentinel that represents a global scope for application commands.
    MENTION_PREFIX _sentinel: A sentinel representing the bot will be mentioned for a prefix
    MISSING _sentinel: A sentinel value that indicates something has not been set

    T TypeVar: A type variable used for generic typing.
    Absent Union[T, Missing]: A type hint for a value that may be MISSING.

"""
import inspect
import logging
import os
import sys
from collections import defaultdict
from importlib.metadata import version as _v
from typing import TypeVar, Union, Callable, Coroutine

__all__ = (
    "__version__",
    "__repo_url__",
    "__py_version__",
    "__api_version__",
    "get_logger",
    "logger_name",
    "kwarg_spam",
    "DISCORD_EPOCH",
    "ACTION_ROW_MAX_ITEMS",
    "SELECTS_MAX_OPTIONS",
    "SELECT_MAX_NAME_LENGTH",
    "CONTEXT_MENU_NAME_LENGTH",
    "SLASH_CMD_NAME_LENGTH",
    "SLASH_CMD_MAX_DESC_LENGTH",
    "SLASH_CMD_MAX_OPTIONS",
    "SLASH_OPTION_NAME_LENGTH",
    "EMBED_MAX_NAME_LENGTH",
    "EMBED_MAX_DESC_LENGTH",
    "EMBED_MAX_FIELDS",
    "EMBED_TOTAL_MAX",
    "EMBED_FIELD_VALUE_LENGTH",
    "Singleton",
    "Sentinel",
    "GlobalScope",
    "Missing",
    "MentionPrefix",
    "GLOBAL_SCOPE",
    "MISSING",
    "MENTION_PREFIX",
    "PREMIUM_GUILD_LIMITS",
    "Absent",
    "T",
    "T_co",
    "LIB_PATH",
    "RECOVERABLE_WEBSOCKET_CLOSE_CODES",
    "NON_RESUMABLE_WEBSOCKET_CLOSE_CODES",
)

_ver_info = sys.version_info

try:
    __version__ = _v("interactions")
except Exception:
    __version__ = "0.0.0"
__repo_url__ = "https://github.com/interactions-py/interactions.py"
__py_version__ = f"{_ver_info[0]}.{_ver_info[1]}"
__api_version__ = 10
logger_name = "interactions"
_logger = logging.getLogger(logger_name)


def get_logger() -> logging.Logger:
    global _logger
    return _logger


default_locale = "english_us"
kwarg_spam = False

DISCORD_EPOCH = 1420070400000

ACTION_ROW_MAX_ITEMS = 5
SELECTS_MAX_OPTIONS = 25
SELECT_MAX_NAME_LENGTH = 100

CONTEXT_MENU_NAME_LENGTH = 32
SLASH_CMD_NAME_LENGTH = 32
SLASH_CMD_MAX_DESC_LENGTH = 100
SLASH_CMD_MAX_OPTIONS = 25
SLASH_OPTION_NAME_LENGTH = 100

EMBED_MAX_NAME_LENGTH = 256
EMBED_MAX_DESC_LENGTH = 4096
EMBED_MAX_FIELDS = 25
EMBED_TOTAL_MAX = 6000
EMBED_FIELD_VALUE_LENGTH = 1024


class Singleton(type):
    _instances = {}

    def __call__(self, *args, **kwargs) -> "Singleton":
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]


class Sentinel(metaclass=Singleton):
    @staticmethod
    def _get_caller_module() -> str:
        stack = inspect.stack()

        caller = stack[2][0]
        return caller.f_globals.get("__name__")

    def __init__(self) -> None:
        self.__module__ = self._get_caller_module()
        self.name = type(self).__name__

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> str:
        return self.name

    def __copy__(self) -> "Sentinel":
        return self

    def __deepcopy__(self, _) -> "Sentinel":
        return self


class GlobalScope(Sentinel, int):
    def __getattr__(self, _) -> "GlobalScope":
        return 0  # type: ignore

    def __hash__(self) -> int:
        return 0

    def __bool__(self) -> bool:
        return False


class Missing(Sentinel):
    def __getattr__(self, *_) -> None:
        return None

    def __bool__(self) -> bool:
        return False


class MentionPrefix(Sentinel):
    ...


GLOBAL_SCOPE = GlobalScope()
MISSING = Missing()
MENTION_PREFIX = MentionPrefix()

PREMIUM_GUILD_LIMITS = defaultdict(
    lambda: {"emoji": 50, "stickers": 5, "bitrate": 96000, "filesize": 8388608},
    {
        1: {"emoji": 100, "stickers": 15, "bitrate": 128000, "filesize": 8388608},
        2: {"emoji": 150, "stickers": 30, "bitrate": 256000, "filesize": 52428800},
        3: {"emoji": 250, "stickers": 60, "bitrate": 384000, "filesize": 104857600},
    },
)


GUILD_WELCOME_MESSAGES = (
    "{0} joined the party.",
    "{0} is here.",
    "Welcome, {0}. We hope you brought pizza.",
    "A wild {0} appeared.",
    "{0} just landed.",
    "{0} just slid into the server.",
    "{0} just showed up!",
    "Welcome {0}. Say hi!",
    "{0} hopped into the server.",
    "Everyone welcome {0}!",
    "Glad you're here, {0}.",
    "Good to see you, {0}.",
    "Yay you made it, {0}!",
)

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
Absent = Union[T, Missing]
AsyncCallable = Callable[..., Coroutine]

LIB_PATH = os.sep.join(__file__.split(os.sep)[:-2])
"""The path to the library folder."""

RECOVERABLE_WEBSOCKET_CLOSE_CODES = (4000, 4001, 4002, 4003, 4005, 4007, 4008, 4009)
NON_RESUMABLE_WEBSOCKET_CLOSE_CODES = (1000, 4007)
